--######################################################################################################
-- FILE          : E_Dx_PAM[n]_Wrap
-- RELATED FILES :
-- DESCRIPTION   : This file is a wrapper file for the Stratix 10 E-Tile PHY in PAM4 mode. 
--               : N, x1 channel PHYs are generated based on the generic entry
--               : 
--               : Different datarates are supported at compile time rather than dynamic reconfiguration
--               :
-- AUTHOR        : James Murray 
-- DEPARTMENT    : European Regional Application Engineering
-- COMPANY       : Intel PSG
--######################################################################################################

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;


entity E_Dx_PAM4_Wrap is
	generic(	NumChannels:integer:=1;
				Rate:string:="USR0";
				SIPatterns:boolean:=TRUE);						
	port(
		refclk				: in std_logic_vector(0 downto 0);
		tx_serial_data		: out std_logic_vector(NumChannels-1 downto 0);
		tx_serial_data_n	: out std_logic_vector(NumChannels-1 downto 0);	
		unusedtx				: out std_logic_vector(NumChannels-1 downto 0);
		unusedtx_n			: out std_logic_vector(NumChannels-1 downto 0);	
				
		rx_serial_data		: in  std_logic_vector(NumChannels-1 downto 0)   := (others => '0'); 
		rx_serial_data_n	: in  std_logic_vector(NumChannels-1 downto 0)   := (others => '0'); 
		unusedrx				: in  std_logic_vector(NumChannels-1 downto 0)   := (others => '0'); 
		unusedrx_n			: in  std_logic_vector(NumChannels-1 downto 0)   := (others => '0'); 
				
		reconfig_clk		: in  std_logic_vector(0 downto 0)   := (others => '0');
		reconfig_reset		: in  std_logic_vector(0 downto 0)   := (others => '0');
		phy_reset			: in  std_logic_vector(0 downto 0)   := (others => '0')		
		);
end entity E_Dx_PAM4_Wrap;

architecture rtl of E_Dx_PAM4_Wrap is

	signal tx_clkout_ch0, tx_clkout_ch1		: std_logic_vector(NumChannels-1 downto 0);
	signal rx_clkout_ch0, rx_clkout_ch1		: std_logic_vector(NumChannels-1 downto 0);
	
	signal tx_ready_ch0        : std_logic_vector(NumChannels-1 downto 0);
	signal rx_ready_ch0        : std_logic_vector(NumChannels-1 downto 0);
	

 	type tx_parallel_data160_array is array (NumChannels-1 downto 0) of std_logic_vector(159 downto 0);  
	signal tx_parallel_data160 : tx_parallel_data160_array; 

	type rx_parallel_data160_array is array (NumChannels-1 downto 0) of std_logic_vector(159 downto 0);  
	signal rx_parallel_data160 : rx_parallel_data160_array; 	
	
 	type tx_parallel_data_ch0_80_array is array (NumChannels-1 downto 0) of std_logic_vector(79 downto 0);  
	signal tx_parallel_data_ch0_80 : tx_parallel_data_ch0_80_array; 

	type rx_parallel_data_ch0_80_array is array (NumChannels-1 downto 0) of std_logic_vector(79 downto 0);  
	signal rx_parallel_data_ch0_80 : rx_parallel_data_ch0_80_array;

component E_Dx_51GPAM is
	PORT(
		reset_ch0                  : in  std_logic                     := 'X';             -- reset_ch0
		tx_ready_ch0               : out std_logic;                                        -- tx_ready_ch0
		rx_ready_ch0               : out std_logic;                                        -- rx_ready_ch0
		tx_pma_ready_ch0           : out std_logic;                                        -- tx_pma_ready
		tx_pma_ready_ch1           : out std_logic;                                        -- tx_pma_ready
		rx_pma_ready_ch0           : out std_logic;                                        -- rx_pma_ready
		rx_pma_ready_ch1           : out std_logic;                                        -- rx_pma_ready
		tx_serial_data_ch0         : out std_logic;                                        -- tx_serial_data
		tx_serial_data_ch1         : out std_logic;                                        -- tx_serial_data
		tx_serial_data_n_ch0       : out std_logic;                                        -- tx_serial_data_n
		tx_serial_data_n_ch1       : out std_logic;                                        -- tx_serial_data_n
		rx_serial_data_ch0         : in  std_logic                     := 'X';             -- rx_serial_data
		rx_serial_data_ch1         : in  std_logic                     := 'X';             -- rx_serial_data
		rx_serial_data_n_ch0       : in  std_logic                     := 'X';             -- rx_serial_data_n
		rx_serial_data_n_ch1       : in  std_logic                     := 'X';             -- rx_serial_data_n
		pll_refclk0                : in  std_logic                     := 'X';             -- clk
		rx_is_lockedtodata_ch0     : out std_logic;                                        -- rx_is_lockedtodata
		rx_is_lockedtodata_ch1     : out std_logic;                                        -- rx_is_lockedtodata
		tx_parallel_data_ch0       : in  std_logic_vector(79 downto 0) := (others => 'X'); -- tx_parallel_data
		tx_parallel_data_ch1       : in  std_logic_vector(79 downto 0) := (others => 'X'); -- tx_parallel_data
		rx_parallel_data_ch0       : out std_logic_vector(79 downto 0);                    -- rx_parallel_data
		rx_parallel_data_ch1       : out std_logic_vector(79 downto 0);                    -- rx_parallel_data
		tx_coreclkin_ch0           : in  std_logic                     := 'X';             -- clk
		tx_coreclkin_ch1           : in  std_logic                     := 'X';             -- clk
		rx_coreclkin_ch0           : in  std_logic                     := 'X';             -- clk
		rx_coreclkin_ch1           : in  std_logic                     := 'X';             -- clk
		tx_clkout_ch0              : out std_logic;                                        -- clk
		tx_clkout_ch1              : out std_logic;                                        -- clk
		rx_clkout_ch0              : out std_logic;                                        -- clk
		rx_clkout_ch1              : out std_logic;                                        -- clk
		reconfig_clk_ch0           : in  std_logic                     := 'X';             -- clk
		reconfig_reset_ch0         : in  std_logic                     := 'X';             -- reset
		reconfig_write_ch0         : in  std_logic                     := 'X';             -- write
		reconfig_read_ch0          : in  std_logic                     := 'X';             -- read
		reconfig_address_ch0       : in  std_logic_vector(19 downto 0) := (others => 'X'); -- address
		reconfig_writedata_ch0     : in  std_logic_vector(7 downto 0)  := (others => 'X'); -- writedata
		reconfig_readdata_ch0      : out std_logic_vector(7 downto 0);                     -- readdata
		reconfig_waitrequest_ch0   : out std_logic;                                        -- waitrequest
		rsfec_avmm2_avmmread_in    : in  std_logic                     := 'X';             -- read
		rsfec_avmm2_avmmrequest_in : in  std_logic                     := 'X';             -- waitrequest
		rsfec_avmm2_avmmwrite_in   : in  std_logic                     := 'X'              -- write
		);
end component;

component E_Dx_USR0PAM is
	PORT(
		reset_ch0                  : in  std_logic                     := 'X';             -- reset_ch0
		tx_ready_ch0               : out std_logic;                                        -- tx_ready_ch0
		rx_ready_ch0               : out std_logic;                                        -- rx_ready_ch0
		tx_pma_ready_ch0           : out std_logic;                                        -- tx_pma_ready
		tx_pma_ready_ch1           : out std_logic;                                        -- tx_pma_ready
		rx_pma_ready_ch0           : out std_logic;                                        -- rx_pma_ready
		rx_pma_ready_ch1           : out std_logic;                                        -- rx_pma_ready
		tx_serial_data_ch0         : out std_logic;                                        -- tx_serial_data
		tx_serial_data_ch1         : out std_logic;                                        -- tx_serial_data
		tx_serial_data_n_ch0       : out std_logic;                                        -- tx_serial_data_n
		tx_serial_data_n_ch1       : out std_logic;                                        -- tx_serial_data_n
		rx_serial_data_ch0         : in  std_logic                     := 'X';             -- rx_serial_data
		rx_serial_data_ch1         : in  std_logic                     := 'X';             -- rx_serial_data
		rx_serial_data_n_ch0       : in  std_logic                     := 'X';             -- rx_serial_data_n
		rx_serial_data_n_ch1       : in  std_logic                     := 'X';             -- rx_serial_data_n
		pll_refclk0                : in  std_logic                     := 'X';             -- clk
		rx_is_lockedtodata_ch0     : out std_logic;                                        -- rx_is_lockedtodata
		rx_is_lockedtodata_ch1     : out std_logic;                                        -- rx_is_lockedtodata
		tx_parallel_data_ch0       : in  std_logic_vector(79 downto 0) := (others => 'X'); -- tx_parallel_data
		tx_parallel_data_ch1       : in  std_logic_vector(79 downto 0) := (others => 'X'); -- tx_parallel_data
		rx_parallel_data_ch0       : out std_logic_vector(79 downto 0);                    -- rx_parallel_data
		rx_parallel_data_ch1       : out std_logic_vector(79 downto 0);                    -- rx_parallel_data
		tx_coreclkin_ch0           : in  std_logic                     := 'X';             -- clk
		tx_coreclkin_ch1           : in  std_logic                     := 'X';             -- clk
		rx_coreclkin_ch0           : in  std_logic                     := 'X';             -- clk
		rx_coreclkin_ch1           : in  std_logic                     := 'X';             -- clk
		tx_clkout_ch0              : out std_logic;                                        -- clk
		tx_clkout_ch1              : out std_logic;                                        -- clk
		rx_clkout_ch0              : out std_logic;                                        -- clk
		rx_clkout_ch1              : out std_logic;                                        -- clk
		reconfig_clk_ch0           : in  std_logic                     := 'X';             -- clk
		reconfig_reset_ch0         : in  std_logic                     := 'X';             -- reset
		reconfig_write_ch0         : in  std_logic                     := 'X';             -- write
		reconfig_read_ch0          : in  std_logic                     := 'X';             -- read
		reconfig_address_ch0       : in  std_logic_vector(19 downto 0) := (others => 'X'); -- address
		reconfig_writedata_ch0     : in  std_logic_vector(7 downto 0)  := (others => 'X'); -- writedata
		reconfig_readdata_ch0      : out std_logic_vector(7 downto 0);                     -- readdata
		reconfig_waitrequest_ch0   : out std_logic;                                        -- waitrequest
		rsfec_avmm2_avmmread_in    : in  std_logic                     := 'X';             -- read
		rsfec_avmm2_avmmrequest_in : in  std_logic                     := 'X';             -- waitrequest
		rsfec_avmm2_avmmwrite_in   : in  std_logic                     := 'X'              -- write
		);
end component;

component E_Dx_USR1PAM is
	PORT(
		reset_ch0                  : in  std_logic                     := 'X';             -- reset_ch0
		tx_ready_ch0               : out std_logic;                                        -- tx_ready_ch0
		rx_ready_ch0               : out std_logic;                                        -- rx_ready_ch0
		tx_pma_ready_ch0           : out std_logic;                                        -- tx_pma_ready
		tx_pma_ready_ch1           : out std_logic;                                        -- tx_pma_ready
		rx_pma_ready_ch0           : out std_logic;                                        -- rx_pma_ready
		rx_pma_ready_ch1           : out std_logic;                                        -- rx_pma_ready
		tx_serial_data_ch0         : out std_logic;                                        -- tx_serial_data
		tx_serial_data_ch1         : out std_logic;                                        -- tx_serial_data
		tx_serial_data_n_ch0       : out std_logic;                                        -- tx_serial_data_n
		tx_serial_data_n_ch1       : out std_logic;                                        -- tx_serial_data_n
		rx_serial_data_ch0         : in  std_logic                     := 'X';             -- rx_serial_data
		rx_serial_data_ch1         : in  std_logic                     := 'X';             -- rx_serial_data
		rx_serial_data_n_ch0       : in  std_logic                     := 'X';             -- rx_serial_data_n
		rx_serial_data_n_ch1       : in  std_logic                     := 'X';             -- rx_serial_data_n
		pll_refclk0                : in  std_logic                     := 'X';             -- clk
		rx_is_lockedtodata_ch0     : out std_logic;                                        -- rx_is_lockedtodata
		rx_is_lockedtodata_ch1     : out std_logic;                                        -- rx_is_lockedtodata
		tx_parallel_data_ch0       : in  std_logic_vector(79 downto 0) := (others => 'X'); -- tx_parallel_data
		tx_parallel_data_ch1       : in  std_logic_vector(79 downto 0) := (others => 'X'); -- tx_parallel_data
		rx_parallel_data_ch0       : out std_logic_vector(79 downto 0);                    -- rx_parallel_data
		rx_parallel_data_ch1       : out std_logic_vector(79 downto 0);                    -- rx_parallel_data
		tx_coreclkin_ch0           : in  std_logic                     := 'X';             -- clk
		tx_coreclkin_ch1           : in  std_logic                     := 'X';             -- clk
		rx_coreclkin_ch0           : in  std_logic                     := 'X';             -- clk
		rx_coreclkin_ch1           : in  std_logic                     := 'X';             -- clk
		tx_clkout_ch0              : out std_logic;                                        -- clk
		tx_clkout_ch1              : out std_logic;                                        -- clk
		rx_clkout_ch0              : out std_logic;                                        -- clk
		rx_clkout_ch1              : out std_logic;                                        -- clk
		reconfig_clk_ch0           : in  std_logic                     := 'X';             -- clk
		reconfig_reset_ch0         : in  std_logic                     := 'X';             -- reset
		reconfig_write_ch0         : in  std_logic                     := 'X';             -- write
		reconfig_read_ch0          : in  std_logic                     := 'X';             -- read
		reconfig_address_ch0       : in  std_logic_vector(19 downto 0) := (others => 'X'); -- address
		reconfig_writedata_ch0     : in  std_logic_vector(7 downto 0)  := (others => 'X'); -- writedata
		reconfig_readdata_ch0      : out std_logic_vector(7 downto 0);                     -- readdata
		reconfig_waitrequest_ch0   : out std_logic;                                        -- waitrequest
		rsfec_avmm2_avmmread_in    : in  std_logic                     := 'X';             -- read
		rsfec_avmm2_avmmrequest_in : in  std_logic                     := 'X';             -- waitrequest
		rsfec_avmm2_avmmwrite_in   : in  std_logic                     := 'X'              -- write
		);
end component;

component DataMux80 is
	PORT(
		Reset								: in  std_logic;
		rx_parallel_datain  : in  std_logic_vector(79 downto 0)    := (others => '0');
		rx_clk							: in  std_logic := '0';
		tx_parallel_dataout : out std_logic_vector(79 downto 0)    := (others => '0');
		tx_clk							: in  std_logic := '0';
		Traffic_Tx_parallel_datain	: in  std_logic_vector(79 downto 0)    := (others => '0')
		);
end component;  

component clk_measure is
	PORT(
		refclock         : in  std_logic; -- clock to be measured
		measure_clk      : in  std_logic; -- fixed reference clock to this module, eg. 100MHz
		reset            : in  std_logic; -- reset signal 
		enable	         : in  std_logic; 
		refclock_measure : out std_logic_vector(23 downto 0);
		valid            : out std_logic
		);
end component;	  
                                                              
begin                                                                                                                                     
                                                           
--xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
-- Instance components here
--xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
	
CH: for i in 0 to NumChannels-1 generate 

	Gen_51GPHY:if Rate = "51G5" generate 
		PHY: E_Dx_51GPAM
			port map(  
				reset_ch0                  => phy_reset(0),							
				tx_ready_ch0               => tx_ready_ch0(i),											
				rx_ready_ch0               => rx_ready_ch0(i),											
				tx_pma_ready_ch0           => open,											
				tx_pma_ready_ch1           => open,											
				rx_pma_ready_ch0           => open,											
				rx_pma_ready_ch1           => open,											
				tx_serial_data_ch0         => tx_serial_data(i),						
				tx_serial_data_ch1         => unusedtx(i),									
				tx_serial_data_n_ch0       => tx_serial_data_n(i),					
				tx_serial_data_n_ch1       => unusedtx_n(i), 							
				rx_serial_data_ch0         => rx_serial_data(i),						
				rx_serial_data_ch1         => unusedrx(i),									
				rx_serial_data_n_ch0       => rx_serial_data_n(i),						
				rx_serial_data_n_ch1       => unusedrx_n(i),								
				pll_refclk0                => refclk(0),								
				rx_is_lockedtodata_ch0     => open,											
				rx_is_lockedtodata_ch1     => open,											
				tx_parallel_data_ch0       => tx_parallel_data_ch0_80(i),
				tx_parallel_data_ch1       => (others => '0'),					
				rx_parallel_data_ch0       => rx_parallel_data_ch0_80(i),
				rx_parallel_data_ch1       => open,											
				tx_coreclkin_ch0           => tx_clkout_ch0(i),
				tx_coreclkin_ch1           => tx_clkout_ch1(i),
				rx_coreclkin_ch0           => rx_clkout_ch0(i),
				rx_coreclkin_ch1           => rx_clkout_ch1(i),
				tx_clkout_ch0              => tx_clkout_ch0(i),
				tx_clkout_ch1              => tx_clkout_ch1(i),
				rx_clkout_ch0              => rx_clkout_ch0(i),
				rx_clkout_ch1              => rx_clkout_ch1(i),
				reconfig_clk_ch0           => reconfig_clk(0),							
				reconfig_reset_ch0         => reconfig_reset(0),									
				reconfig_write_ch0         => '0',					
				reconfig_read_ch0          => '0',					
				reconfig_address_ch0       => (others => '0'),					
				reconfig_writedata_ch0     => (others => '0'),					
				reconfig_readdata_ch0      => open,											
				reconfig_waitrequest_ch0   => open,											
				rsfec_avmm2_avmmread_in    => '0',											
				rsfec_avmm2_avmmrequest_in => '0',											
				rsfec_avmm2_avmmwrite_in   => '0'												
				);
	end generate Gen_51GPHY;
	
	Gen_USR0PHY:if Rate = "USR0" generate 
		PHY: E_Dx_USR0PAM
			port map(   
				reset_ch0                  => phy_reset(0),							
				tx_ready_ch0               => tx_ready_ch0(i),											
				rx_ready_ch0               => rx_ready_ch0(i),											
				tx_pma_ready_ch0           => open,											
				tx_pma_ready_ch1           => open,											
				rx_pma_ready_ch0           => open,											
				rx_pma_ready_ch1           => open,											
				tx_serial_data_ch0         => tx_serial_data(i),						
				tx_serial_data_ch1         => unusedtx(i),									
				tx_serial_data_n_ch0       => tx_serial_data_n(i),					
				tx_serial_data_n_ch1       => unusedtx_n(i), 							
				rx_serial_data_ch0         => rx_serial_data(i),						
				rx_serial_data_ch1         => unusedrx(i),									
				rx_serial_data_n_ch0       => rx_serial_data_n(i),						
				rx_serial_data_n_ch1       => unusedrx_n(i),								
				pll_refclk0                => refclk(0),								
				rx_is_lockedtodata_ch0     => open,											
				rx_is_lockedtodata_ch1     => open,											
				tx_parallel_data_ch0       => tx_parallel_data_ch0_80(i),
				tx_parallel_data_ch1       => (others => '0'),					
				rx_parallel_data_ch0       => rx_parallel_data_ch0_80(i),
				rx_parallel_data_ch1       => open,											
				tx_coreclkin_ch0           => tx_clkout_ch0(i),
				tx_coreclkin_ch1           => tx_clkout_ch1(i),
				rx_coreclkin_ch0           => rx_clkout_ch0(i),
				rx_coreclkin_ch1           => rx_clkout_ch1(i),
				tx_clkout_ch0              => tx_clkout_ch0(i),
				tx_clkout_ch1              => tx_clkout_ch1(i),
				rx_clkout_ch0              => rx_clkout_ch0(i),
				rx_clkout_ch1              => rx_clkout_ch1(i),
				reconfig_clk_ch0           => reconfig_clk(0),							
				reconfig_reset_ch0         => reconfig_reset(0),									
				reconfig_write_ch0         => '0',					
				reconfig_read_ch0          => '0',					
				reconfig_address_ch0       => (others => '0'),					
				reconfig_writedata_ch0     => (others => '0'),					
				reconfig_readdata_ch0      => open,											
				reconfig_waitrequest_ch0   => open,											
				rsfec_avmm2_avmmread_in    => '0',											
				rsfec_avmm2_avmmrequest_in => '0',											
				rsfec_avmm2_avmmwrite_in   => '0'											
				);
	end generate Gen_USR0PHY;
			
	Gen_USR1PHY:if Rate = "USR1" generate 
		PHY: E_Dx_USR1PAM
			port map(   
				reset_ch0                  => phy_reset(0),							
				tx_ready_ch0               => tx_ready_ch0(i),											
				rx_ready_ch0               => rx_ready_ch0(i),											
				tx_pma_ready_ch0           => open,											
				tx_pma_ready_ch1           => open,											
				rx_pma_ready_ch0           => open,											
				rx_pma_ready_ch1           => open,											
				tx_serial_data_ch0         => tx_serial_data(i),						
				tx_serial_data_ch1         => unusedtx(i),									
				tx_serial_data_n_ch0       => tx_serial_data_n(i),					
				tx_serial_data_n_ch1       => unusedtx_n(i), 							
				rx_serial_data_ch0         => rx_serial_data(i),						
				rx_serial_data_ch1         => unusedrx(i),									
				rx_serial_data_n_ch0       => rx_serial_data_n(i),						
				rx_serial_data_n_ch1       => unusedrx_n(i),								
				pll_refclk0                => refclk(0),								
				rx_is_lockedtodata_ch0     => open,											
				rx_is_lockedtodata_ch1     => open,											
				tx_parallel_data_ch0       => tx_parallel_data_ch0_80(i),
				tx_parallel_data_ch1       => (others => '0'),					
				rx_parallel_data_ch0       => rx_parallel_data_ch0_80(i),
				rx_parallel_data_ch1       => open,											
				tx_coreclkin_ch0           => tx_clkout_ch0(i),
				tx_coreclkin_ch1           => tx_clkout_ch1(i),
				rx_coreclkin_ch0           => rx_clkout_ch0(i),
				rx_coreclkin_ch1           => rx_clkout_ch1(i),
				tx_clkout_ch0              => tx_clkout_ch0(i),
				tx_clkout_ch1              => tx_clkout_ch1(i),
				rx_clkout_ch0              => rx_clkout_ch0(i),
				rx_clkout_ch1              => rx_clkout_ch1(i),
				reconfig_clk_ch0           => reconfig_clk(0),							
				reconfig_reset_ch0         => reconfig_reset(0),									
				reconfig_write_ch0         => '0',					
				reconfig_read_ch0          => '0',					
				reconfig_address_ch0       => (others => '0'),					
				reconfig_writedata_ch0     => (others => '0'),					
				reconfig_readdata_ch0      => open,											
				reconfig_waitrequest_ch0   => open,											
				rsfec_avmm2_avmmread_in    => '0',											
				rsfec_avmm2_avmmrequest_in => '0',											
				rsfec_avmm2_avmmwrite_in   => '0'											
				);
	end generate Gen_USR1PHY;	
	
	Gen_Mux:if SIPatterns = TRUE generate 		
		DataPathSel: DataMux80
			port map(   
				Reset									=> tx_ready_ch0(i),
				rx_parallel_datain   			=> rx_parallel_data_ch0_80(i),
				rx_clk								=> rx_clkout_ch0(i),
				tx_parallel_dataout				=> tx_parallel_data_ch0_80(i),
				tx_clk								=> tx_clkout_ch0(i),
				Traffic_Tx_parallel_datain		=> x"BCBC" & x"BCBC" & x"BCBC" & x"BCBC" & x"BCBC"
				);
	end generate Gen_Mux;			

end generate; 

measure_txclkout0_i:clk_measure
	port map(
		refclock         => tx_clkout_ch0(0),						--: in  std_logic; // clock to be measured
		measure_clk      => reconfig_clk(0),				--: in  std_logic; // fixed reference clock to this module, eg. 100MHz
		reset            => reconfig_reset(0),			--: in  std_logic; // reset signal 
		enable	         => '1',
		refclock_measure => open,
		valid            => open
		);
		
measure_rxclkout0_i:clk_measure
	port map(
		refclock         => rx_clkout_ch0(0),						--: in  std_logic; // clock to be measured
		measure_clk      => reconfig_clk(0),				--: in  std_logic; // fixed reference clock to this module, eg. 100MHz
		reset            => reconfig_reset(0),			--: in  std_logic; // reset signal 
		enable	         => '1',
		refclock_measure => open,
		valid            => open
		);
    	

end rtl;
